class BaseCalibrator(object):
    def __str__(self):
        raise NotImplementedError()
    
    def __copy__(self):
        raise NotImplementedError()

    def fit(self, logits, labels):
        raise NotImplementedError()

    def calibrate(self):
        raise NotImplementedError()



from scipy.optimize import minimize
from sklearn.metrics import log_loss

from labelshift.utils import softmax

from .base import BaseCalibrator


class TemperatureScaling(BaseCalibrator):
    def __init__(self, temp=1, maxiter=100, solver="BFGS"):
        self.temp = temp
        self.maxiter = maxiter
        self.solver = solver
    
    def __str__(self):
        return "Temperature Scaling (TS)"

    def __copy__(self):
        return TemperatureScaling(self.temp, self.maxiter, self.solver)

    def _loss_func(self, x, probs, labels):
        scaled_probs = self.calibrate(probs, x)
        loss = log_loss(y_true=labels, y_pred=scaled_probs)
        return loss

    def calibrate(self, logits, temp=None):
        if temp is not None:
            return softmax(logits / temp)
        else:
            return softmax(logits / self.temp)

    def fit(self, logits, labels):
        init_guess = 1.0

        result = minimize(
            fun=self._loss_func,
            args=(logits, labels),
            x0=init_guess,
            method=self.solver,
            tol=1e-07,
            options={"maxiter": self.maxiter},
        )

        assert result.success == True, f"TS Optimization Failed:\n{result}"
        self.temp = result.x[0]
        return self



from labelshift.methods import *
from labelshift.calibrations import *


def get_lse_methods(lse_algs, calibrations, use_ensemble=True):
    names, estimators = [], []
    for lse_alg in lse_algs:
        for calibration in calibrations:
            names.append(f"{lse_alg}_{calibration}")
            estimator = get_lse_method(lse_alg, calibration, use_ensemble)
            estimators.append(estimator)

    return names, estimators


def get_lse_method(lse_alg, calibration=None, use_ensemble=True):
    # get calibration method
    if calibration is None or calibration.upper() == "NONE":
        calibrator = None
    elif calibration.upper() == "TS":
        calibrator = TemperatureScaling()
    elif calibration.upper() == "BCTS":
        calibrator = BCTemperatureScaling()
    else:
        raise RuntimeError(f"calibration method: {calibration} is not implemented.")

    # get label shift estimation method
    if lse_alg.upper() == "BBSE":
        Estimator = BBSE
    elif lse_alg.upper() == "RLLS":
        Estimator = RLLS
    elif lse_alg.upper() == "MLLS":
        Estimator = MLLS
    else:
        raise RuntimeError(f"label shift estimator: {lse_alg} is not implemented.")

    estimator = Estimator(calibrator=calibrator)

    if use_ensemble:
        estimator = EnsembleWrapper(estimator)

    return estimator



import numpy as np


class BaseEstimator:
    def __str__(self):
        raise NotImplementedError()
    
    def __copy__(self):
        raise NotImplementedError()

    def fit(self, target_logits, src_logits, src_labels):
        raise NotImplementedError()

    def calibrate(self, logits):
        if self.calibrator is not None:
            logits = self.calibrator.calibrate(logits)
        cal_logits = logits * self.imp_weights[None, :]
        cal_logits = cal_logits / np.sum(cal_logits, axis=-1)[:, None]
        return cal_logits



import cvxpy as cp
import numpy as np

from labelshift.utils import to_one_hot, softmax, to_hard_labels, get_confusion_matrix

from .base import BaseEstimator


class RLLS(BaseEstimator):
    def __init__(self, hard_label=False, calibrator=None, alpha=0.01, verbose=False):
        self.hard_label = hard_label
        self.calibrator = calibrator
        self.alpha = alpha
        self.verbose = verbose
        self.estim_source_dist = None  # estimate source class distribution
        self.estim_target_dist = None  # estimate target class distribution
        self.imp_weights = None  # importance weights

    def __str__(self):
        return "Regularized Learning of Label Shift (RLLS)"

    def __copy__(self):
        calirator_copy = self.calibrator.__copy__() if self.calibrator is not None else None
        rlls_copy = RLLS(self.hard_label, calirator_copy, self.alpha, self.verbose)
        rlls_copy.estim_source_dist = self.estim_source_dist
        rlls_copy.estim_target_dist = self.estim_target_dist
        rlls_copy.imp_weights = self.imp_weights
        return rlls_copy

    # modified from: https://github.com/Angela0428/labelshift/blob/5bbe517938f4e3f5bd14c2c105de973dcc2e0917/label_shift.py
    def compute_3deltaC(self, n_class, n_train, delta):
        rho = 3 * (2 * np.log(2 * n_class / delta) / (3 * n_train) + np.sqrt(2 * np.log(2 * n_class / delta) / n_train))
        return rho

    # modified from: https://github.com/Angela0428/labelshift/blob/5bbe517938f4e3f5bd14c2c105de973dcc2e0917/label_shift.py
    def compute_w_opt(self, C_yy, mu_y, mu_train_y, rho):
        n = C_yy.shape[0]
        theta = cp.Variable(n)
        b = mu_y - mu_train_y
        objective = cp.Minimize(cp.pnorm(C_yy @ theta - b) + rho * cp.pnorm(theta))
        constraints = [-1 <= theta]
        prob = cp.Problem(objective, constraints)

        # The optimal objective value is returned by `prob.solve()`.
        result = prob.solve()
        # The optimal value for x is stored in `x.value`.
        # print(theta.value)
        w = 1 + theta.value
        return w

    def fit(self, target_logits, src_logits, src_labels):
        src_labels = to_one_hot(src_labels)

        if self.calibrator is not None:
            assert src_labels is not None, "source_labels can't be None"
            self.calibrator.fit(src_logits, src_labels)
            src_logits = self.calibrator.calibrate(src_logits)
            target_logits = self.calibrator.calibrate(target_logits)
        else:
            src_logits = softmax(src_logits)
            target_logits = softmax(target_logits)

        if self.hard_label:
            target_logits = to_hard_labels(target_logits, one_hot=True)
            src_logits = to_hard_labels(src_logits, one_hot=True)

        self.estim_source_dist = np.mean(src_labels, axis=0)
        self.estim_source_dist /= np.sum(self.estim_source_dist)  # normalize

        mean_target_logits = np.mean(target_logits, axis=0)
        mean_src_logits = np.mean(src_logits, axis=0)
        confusion_matrix = get_confusion_matrix(src_labels, src_logits, normalize=True)

        num_train = src_logits.shape[0]
        num_classes = src_logits.shape[1]
        rho = self.compute_3deltaC(num_classes, num_train, delta=0.05)
        self.imp_weights = self.compute_w_opt(
            C_yy=confusion_matrix,
            mu_y=mean_target_logits,
            mu_train_y=mean_src_logits,
            rho=self.alpha * rho,
        )
        self.imp_weights = self.imp_weights * (self.imp_weights >= 0)  # mask out negative weights

        self.estim_target_dist = self.imp_weights * self.estim_source_dist
        self.estim_target_dist = self.estim_target_dist / np.sum(self.estim_target_dist)
        self.imp_weights = self.estim_target_dist / self.estim_source_dist

        if self.verbose:
            with np.printoptions(precision=3, suppress=True, formatter={"float": "{: 0.3f}".format}):
                print(f"Est Source Class Distribution: {self.estim_source_dist}")
                print(f"Est Target Class Distribution: {self.estim_target_dist}")
                print(f"Est Importance Weights       : {self.imp_weights}")

        return self



import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F


class Get_Scalar:
    def __init__(self, value):
        self.value = value

    def get_value(self, iter):
        return self.value

    def __call__(self, iter):
        return self.value


class DRWLoss:
    def __init__(self, args, class_num_list, warm_iter=100000, beta=0.9999):
        self.args = args
        self.beta = beta
        self.warm_iter = warm_iter
        self.uniform_weights = self.get_per_class_weights(0, class_num_list).cuda(args.gpu)
        self.per_class_weights = self.get_per_class_weights(self.beta, class_num_list).cuda(args.gpu)

    def get_per_class_weights(self, beta, class_num_list):
        effective_num = 1.0 - np.power(beta, class_num_list)
        per_class_weights = (1.0 - beta) / np.array(effective_num)
        per_class_weights = per_class_weights / np.sum(per_class_weights) * len(class_num_list)
        return torch.FloatTensor(per_class_weights)

    def update_dist(self, class_num_list):
        self.per_class_weights = self.get_per_class_weights(self.beta, class_num_list).cuda(self.args.gpu)

    def __call__(self, it, logits, targets, use_hard_labels=True, reduction="none"):
        weight = self.uniform_weights if it < self.warm_iter else self.per_class_weights
        if use_hard_labels:
            log_pred = F.log_softmax(logits, dim=-1)
            return F.nll_loss(log_pred, targets, weight=weight, reduction=reduction)
        else:
            assert logits.shape == targets.shape
            log_pred = F.log_softmax(logits, dim=-1)
            nll_loss = torch.sum(-targets * log_pred * weight, dim=1)
            return nll_loss


def mixup_one_target(x, y, gpu, alpha=1.0, is_bias=False):
    """Returns mixed inputs, mixed targets, and lambda"""
    if alpha > 0:
        lam = np.random.beta(alpha, alpha)
    else:
        lam = 1
    if is_bias:
        lam = max(lam, 1 - lam)

    index = torch.randperm(x.size(0)).cuda(gpu)

    mixed_x = lam * x + (1 - lam) * x[index, :]
    mixed_y = lam * y + (1 - lam) * y[index]
    return mixed_x, mixed_y, lam


def to_one_hot(targets, nClass, gpu):
    logits = torch.zeros(targets.size(0), nClass).cuda(gpu)
    return logits.scatter_(1, targets.unsqueeze(1), 1)


def ce_loss(logits, targets, use_hard_labels=True, reduction="none"):
    """
    wrapper for cross entropy loss in pytorch.

    Args
        logits: logit values, shape=[Batch size, # of classes]
        targets: integer or vector, shape=[Batch size] or [Batch size, # of classes]
        use_hard_labels: If True, targets have [Batch size] shape with int values. If False, the target is vector (default True)
    """
    if use_hard_labels:
        log_pred = F.log_softmax(logits, dim=-1)
        return F.nll_loss(log_pred, targets, reduction=reduction)
        # return F.cross_entropy(logits, targets, reduction=reduction) this is unstable
    else:
        assert logits.shape == targets.shape
        log_pred = F.log_softmax(logits, dim=-1)
        nll_loss = torch.sum(-targets * log_pred, dim=1)
        return nll_loss


class EMA:
    """
    Implementation from https://fyubang.com/2019/06/01/ema/
    """

    def __init__(self, model, decay):
        self.model = model
        self.decay = decay
        self.shadow = {}
        self.backup = {}

    def load(self, ema_model):
        for name, param in ema_model.named_parameters():
            self.shadow[name] = param.data.clone()

    def register(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                self.shadow[name] = param.data.clone()

    def update(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                assert name in self.shadow
                new_average = (1.0 - self.decay) * param.data + self.decay * self.shadow[name]
                self.shadow[name] = new_average.clone()

    def apply_shadow(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                assert name in self.shadow
                self.backup[name] = param.data
                param.data = self.shadow[name]

    def restore(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                assert name in self.backup
                param.data = self.backup[name]
        self.backup = {}


class Bn_Controller:
    def __init__(self):
        """
        freeze_bn and unfreeze_bn must appear in pairs
        """
        self.backup = {}

    def freeze_bn(self, model):
        assert self.backup == {}
        for name, m in model.named_modules():
            if isinstance(m, nn.SyncBatchNorm) or isinstance(m, nn.BatchNorm2d):
                self.backup[name + ".running_mean"] = m.running_mean.data.clone()
                self.backup[name + ".running_var"] = m.running_var.data.clone()
                self.backup[name + ".num_batches_tracked"] = m.num_batches_tracked.data.clone()

    def unfreeze_bn(self, model):
        for name, m in model.named_modules():
            if isinstance(m, nn.SyncBatchNorm) or isinstance(m, nn.BatchNorm2d):
                m.running_mean.data = self.backup[name + ".running_mean"]
                m.running_var.data = self.backup[name + ".running_var"]
                m.num_batches_tracked.data = self.backup[name + ".num_batches_tracked"]
        self.backup = {}



import numpy as np


class EnsembleWrapper:
    def __init__(self, estimator):
        self.estimator = estimator
        self.estimators = []
        self.estim_source_dist = None
        self.estim_target_dist = None
        self.imp_weights = None

    def __str__(self):
        return self.estimator.__str__()

    def fit(self, target_logits, src_logits, src_labels=None):
        """
        Args:
            target_logits (list): list of target logits
            src_logits (list): list of source logits
            src_labels (list): list of source labels
        """
        if src_labels is None:
            src_labels = [None] * len(src_logits)
        assert len(target_logits) == len(src_logits) == len(src_labels), "length of target_logits, src_logits and src_labels mismatch"

        estim_source_dists = []
        estim_target_dists = []
        imp_weights = []
        for _target_logits, _src_logits, _src_labels in zip(target_logits, src_logits, src_labels):
            _estimator = self.estimator.__copy__()
            e = _estimator.fit(_target_logits, _src_logits, _src_labels)
            estim_source_dists.append(e.estim_source_dist)
            estim_target_dists.append(e.estim_target_dist)
            imp_weights.append(e.imp_weights)
            self.estimators.append(_estimator)

        self.estim_source_dist = np.mean(estim_source_dists, axis=0)
        self.estim_target_dist = np.mean(estim_target_dists, axis=0)
        self.imp_weights = np.mean(imp_weights, axis=0)

        return self

    def calibrate(self, logits):
        assert len(logits) == len(self.estimators)
        outputs = []
        for _logits, _estimator in zip(logits, self.estimators):
            output = _estimator.calibrate(_logits)
            outputs.append(output)
        return np.mean(outputs, axis=0)



"""
source: https://github.com/TorchSSL/TorchSSL
"""
import torch
import torch.nn as nn
import torch.nn.functional as F

momentum = 0.001


def mish(x):
    """Mish: A Self Regularized Non-Monotonic Neural Activation Function (https://arxiv.org/abs/1908.08681)"""
    return x * torch.tanh(F.softplus(x))


class PSBatchNorm2d(nn.BatchNorm2d):
    """How Does BN Increase Collapsed Neural Network Filters? (https://arxiv.org/abs/2001.11216)"""

    def __init__(self, num_features, alpha=0.1, eps=1e-05, momentum=0.001, affine=True, track_running_stats=True):
        super().__init__(num_features, eps, momentum, affine, track_running_stats)
        self.alpha = alpha

    def forward(self, x):
        return super().forward(x) + self.alpha


class BasicBlock(nn.Module):
    def __init__(self, in_planes, out_planes, stride, drop_rate=0.0, activate_before_residual=False):
        super(BasicBlock, self).__init__()
        self.bn1 = nn.BatchNorm2d(in_planes, momentum=0.001, eps=0.001)
        self.relu1 = nn.LeakyReLU(negative_slope=0.1, inplace=False)
        self.conv1 = nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=True)
        self.bn2 = nn.BatchNorm2d(out_planes, momentum=0.001, eps=0.001)
        self.relu2 = nn.LeakyReLU(negative_slope=0.1, inplace=False)
        self.conv2 = nn.Conv2d(out_planes, out_planes, kernel_size=3, stride=1, padding=1, bias=True)
        self.drop_rate = drop_rate
        self.equalInOut = in_planes == out_planes
        self.convShortcut = (
            (not self.equalInOut) and nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, padding=0, bias=True) or None
        )
        self.activate_before_residual = activate_before_residual

    def forward(self, x):
        if not self.equalInOut and self.activate_before_residual == True:
            x = self.relu1(self.bn1(x))
        else:
            out = self.relu1(self.bn1(x))
        out = self.relu2(self.bn2(self.conv1(out if self.equalInOut else x)))
        if self.drop_rate > 0:
            out = F.dropout(out, p=self.drop_rate, training=self.training)
        out = self.conv2(out)
        return torch.add(x if self.equalInOut else self.convShortcut(x), out)


class NetworkBlock(nn.Module):
    def __init__(self, nb_layers, in_planes, out_planes, block, stride, drop_rate=0.0, activate_before_residual=False):
        super(NetworkBlock, self).__init__()
        self.layer = self._make_layer(block, in_planes, out_planes, nb_layers, stride, drop_rate, activate_before_residual)

    def _make_layer(self, block, in_planes, out_planes, nb_layers, stride, drop_rate, activate_before_residual):
        layers = []
        for i in range(int(nb_layers)):
            layers.append(
                block(i == 0 and in_planes or out_planes, out_planes, i == 0 and stride or 1, drop_rate, activate_before_residual)
            )
        return nn.Sequential(*layers)

    def forward(self, x):
        return self.layer(x)


class WideResNet(nn.Module):
    def __init__(self, first_stride, num_classes, depth=28, widen_factor=2, drop_rate=0.0):
        super(WideResNet, self).__init__()
        channels = [16, 16 * widen_factor, 32 * widen_factor, 64 * widen_factor]
        assert (depth - 4) % 6 == 0
        n = (depth - 4) / 6
        block = BasicBlock
        # 1st conv before any network block
        self.conv1 = nn.Conv2d(3, channels[0], kernel_size=3, stride=1, padding=1, bias=True)
        # 1st block
        self.block1 = NetworkBlock(n, channels[0], channels[1], block, first_stride, drop_rate, activate_before_residual=True)
        # 2nd block
        self.block2 = NetworkBlock(n, channels[1], channels[2], block, 2, drop_rate)
        # 3rd block
        self.block3 = NetworkBlock(n, channels[2], channels[3], block, 2, drop_rate)
        # global average pooling and classifier
        self.bn1 = nn.BatchNorm2d(channels[3], momentum=0.001, eps=0.001)
        self.relu = nn.LeakyReLU(negative_slope=0.1, inplace=False)
        self.fc = nn.Linear(channels[3], num_classes)
        self.channels = channels[3]

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="leaky_relu")
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                nn.init.xavier_normal_(m.weight.data)
                m.bias.data.zero_()

    def forward(self, x, ood_test=False):
        out = self.conv1(x)
        out = self.block1(out)
        out = self.block2(out)
        out = self.block3(out)
        out = self.relu(self.bn1(out))
        out = F.adaptive_avg_pool2d(out, 1)
        out = out.view(-1, self.channels)
        output = self.fc(out)

        if ood_test:
            return output, out
        else:
            return output


class build_WideResNet:
    def __init__(self, first_stride=1, depth=28, widen_factor=2, bn_momentum=0.01, leaky_slope=0.0, dropRate=0.0, use_embed=False):
        self.first_stride = first_stride
        self.depth = depth
        self.widen_factor = widen_factor
        self.bn_momentum = bn_momentum
        self.dropRate = dropRate
        self.leaky_slope = leaky_slope
        self.use_embed = use_embed

    def build(self, num_classes):
        return WideResNet(
            first_stride=self.first_stride,
            depth=self.depth,
            num_classes=num_classes,
            widen_factor=self.widen_factor,
            drop_rate=self.dropRate,
        )


if __name__ == "__main__":
    wrn_builder = build_WideResNet(1, 10, 2, 0.01, 0.1, 0.5)
    wrn = wrn_builder.build(10)
    print(wrn)



import numpy as np
from numpy.random import dirichlet

import torch
from torch.utils.data import sampler, DataLoader
from torch.utils.data.sampler import BatchSampler


def sample_data(data, target, class_num_list, replace=False):
    """
    FIXME: add description
    samples for labeled data
    (sampling with imbalanced ratio over classes)
    """
    lb_data, lbs, lb_idx = [], [], []
    for c in range(len(class_num_list)):
        idx = np.where(target == c)[0]
        assert replace or len(idx) >= class_num_list[c], "insufficient unlabeled data to select"
        idx = np.random.choice(idx, class_num_list[c], replace=replace)
        lb_idx.extend(idx)

        lb_data.extend(data[idx])
        lbs.extend(target[idx])

    return np.array(lb_data), np.array(lbs), np.array(lb_idx)


def gen_imb_list(max_num_cls_lb, imb_ratio, imb_type, num_classes):
    # FIXME: add description
    imb_ratio_lt1 = imb_ratio < 1
    if imb_ratio_lt1:
        imb_ratio = 1 / imb_ratio

    if imb_type == "long":
        mu = np.power(1 / imb_ratio, 1 / (num_classes - 1))
        class_num_list = []
        for i in range(num_classes):
            if i == (num_classes - 1):
                class_num_list.append(int(max_num_cls_lb / imb_ratio))
            else:
                class_num_list.append(int(max_num_cls_lb * np.power(mu, i)))

    if imb_type == "step":
        class_num_list = []
        for i in range(num_classes):
            if i < int(num_classes / 2):
                class_num_list.append(int(max_num_cls_lb))
            else:
                class_num_list.append(int(max_num_cls_lb / imb_ratio))

    if imb_ratio_lt1:
        class_num_list.reverse()

    class_num_list = np.array(class_num_list).astype(int)

    return class_num_list


def gen_dirichlet_list(num_data, alpha, num_classes):
    # FIXME: add description
    if alpha == "uniform":
        dist = np.ones(num_classes) / num_classes
    else:
        dist = dirichlet([alpha] * num_classes)

    class_num_list = dist * num_data
    class_num_list = class_num_list.round().astype(int)

    return class_num_list


def get_sampler_by_name(name):
    """
    get sampler in torch.utils.data.sampler by name
    """
    sampler_name_list = sorted(
        name for name in torch.utils.data.sampler.__dict__ if not name.startswith("_") and callable(sampler.__dict__[name])
    )
    try:
        if name == "RandomSampler":
            from labelshift.datasets.RandomSampler import RandomSampler

            return RandomSampler
        else:
            return getattr(torch.utils.data.sampler, name)
    except Exception as e:
        print(repr(e))
        print("[!] select sampler in:\t", sampler_name_list)


def get_data_loader(
    dset,
    batch_size=None,
    shuffle=False,
    num_workers=4,
    pin_memory=False,
    data_sampler=None,
    replacement=True,
    num_epochs=None,
    num_iters=None,
    generator=None,
    drop_last=True,
):
    """
    get_data_loader returns torch.utils.data.DataLoader for a Dataset.
    All arguments are comparable with those of pytorch DataLoader.
    However, if distributed, DistributedProxySampler, which is a wrapper of data_sampler, is used.

    Args
        num_epochs: total batch -> (# of batches in dset) * num_epochs
        num_iters: total batch -> num_iters
    """

    assert batch_size is not None

    if data_sampler is None:
        return DataLoader(dset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, pin_memory=pin_memory)

    else:
        if isinstance(data_sampler, str):
            data_sampler = get_sampler_by_name(data_sampler)

        num_replicas = 1

        if (num_epochs is not None) and (num_iters is None):
            num_samples = len(dset) * num_epochs
        elif (num_epochs is None) and (num_iters is not None):
            num_samples = batch_size * num_iters * num_replicas
        else:
            num_samples = len(dset)

        if data_sampler.__name__ == "RandomSampler":
            data_sampler = data_sampler(dset, replacement, num_samples, generator)
        else:
            raise RuntimeError(f"{data_sampler.__name__} is not implemented.")

        batch_sampler = BatchSampler(data_sampler, batch_size, drop_last)
        return DataLoader(dset, batch_sampler=batch_sampler, num_workers=num_workers, pin_memory=pin_memory)


def get_onehot(num_classes, idx):
    onehot = np.zeros([num_classes], dtype=np.float32)
    onehot[idx] += 1.0
    return onehot



from scipy.optimize import minimize
from sklearn.metrics import log_loss

from labelshift.utils import softmax

from .base import BaseCalibrator


# Bias Corrected Temperature Scaling
class BCTemperatureScaling(BaseCalibrator):
    def __init__(self, temp=1, bias=0, maxiter=100, solver="L-BFGS-B"):
        self.temp = temp
        self.biases = bias
        self.maxiter = maxiter
        self.solver = solver

    def __str__(self):
        return "Bias-Corrected Temperature Scaling (BCTS)"

    def __copy__(self):
        return BCTemperatureScaling(self.temp, self.biases, self.maxiter, self.solver)

    def _loss_func(self, x, probs, labels):
        temp, biases = x[0], x[1:]
        scaled_probs = self.calibrate(probs, temp, biases)
        loss = log_loss(y_true=labels, y_pred=scaled_probs)
        return loss

    def calibrate(self, logits, temp=None, biases=None):
        if temp is not None and biases is not None:
            return softmax(logits / temp + biases)
        else:
            return softmax(logits / self.temp + self.biases)

    def fit(self, logits, labels):
        init_guess = [1.0] + [0.0 for _ in range(logits.shape[1])]

        result = minimize(
            fun=self._loss_func,
            args=(logits, labels),
            x0=init_guess,
            method=self.solver,
            tol=1e-07,
            options={"maxiter": self.maxiter},
        )

        assert result.success == True, f"BCTS Optimization Failed:\n{result}"
        self.temp, self.biases = result.x[0], result.x[1:]
        return self



from .imbalanced_dataset import Imbalanced_Dataset
from .dirichlet_dataset import Dirichlet_Dataset
from .data_utils import get_data_loader
from .dataset import BasicDataset, ResampleDataset, ResampleDataset

__all__ = ["Imbalanced_Dataset", "Dirichlet_Dataset", "get_data_loader", "BasicDataset", "ResampleDataset", "ResampleDataset"]



# import needed library
import json
import copy
import random
import shutil
import logging
import warnings
import numpy as np
from pathlib import Path
from sklearn.metrics import mean_squared_error

import torch
import torch.nn.parallel
import torch.backends.cudnn as cudnn
from torchvision import transforms

from labelshift.lse.config import Config
from labelshift.model import ModelTrainer
from labelshift.get_method import get_lse_methods
from labelshift.datasets import get_data_loader, BasicDataset, ResampleDataset, ResampleDataset
from labelshift.utils import net_builder, get_optimizer, get_cosine_schedule_with_warmup, labels_to_dist, over_write_args_from_dict


class LSE:
    def __init__(self, args=None, **kwargs):
        if args is not None:
            self.cfg = over_write_args_from_dict(args, kwargs)
        else:
            self.cfg = Config(**kwargs)
        
        print(f"\n{self.cfg}\n")
        if Path(self.cfg.save_path).exists() and self.cfg.overwrite:
            shutil.rmtree(self.cfg.save_path)
    
    def train_base_models(self, lb_data, lb_targets, num_classes=None, train_transform=None, test_transform=None):
        self.pre_check()

        if num_classes is not None:
            self.cfg.num_classes = num_classes
        assert self.cfg.num_classes > 0, "please provide the number of classes"
        
        train_transform = train_transform if train_transform is not None else transforms.ToTensor()
        test_transform = test_transform if test_transform is not None else transforms.ToTensor()

        random_state = random.getstate()
        np_random_state = np.random.get_state()
        torch_random_state = torch.random.get_rng_state()
        
        if self.cfg.seed is not None:
            warnings.warn(
                "You have chosen to seed training. "
                "This will turn on the CUDNN deterministic setting, "
                "which can slow down your training considerably! "
                "You may see unexpected behavior when restarting "
                "from checkpoints.\n"
            )
            random.seed(self.cfg.seed)
            torch.manual_seed(self.cfg.seed)
            np.random.seed(self.cfg.seed)
            cudnn.deterministic = True

        cudnn.benchmark = True

        # Construct Dataset & DataLoader
        lb_dset = ResampleDataset(lb_data, lb_targets, self.cfg.num_classes, train_transform, test_transform, onehot=False)
        
        _net_builder = net_builder(
            self.cfg.net,
            self.cfg.net_from_name,
            {
                "first_stride": 2 if "stl" in self.cfg.dataset else 1,
                "depth": self.cfg.depth,
                "widen_factor": self.cfg.widen_factor,
                "leaky_slope": self.cfg.leaky_slope,
                "bn_momentum": self.cfg.bn_momentum,
                "dropRate": self.cfg.dropout,
                "use_embed": False,
            },
        )
        print()

        for idx, train_lb_dset, val_lb_dset in lb_dset.resample(self.cfg.num_val_per_class, self.cfg.num_ensemble, seed=self.cfg.seed):
            print(f"Training [{idx}/{self.cfg.num_ensemble}] Model")
            model = ModelTrainer(_net_builder, self.cfg.num_classes, num_eval_iter=self.cfg.num_eval_iter, ema_m=self.cfg.ema_m)
            # SET Optimizer & LR Scheduler
            ## construct SGD and cosine lr scheduler
            optimizer = get_optimizer(model.model, self.cfg.optim, self.cfg.lr, self.cfg.momentum, self.cfg.weight_decay)
            scheduler = get_cosine_schedule_with_warmup(optimizer, self.cfg.num_train_iter, num_warmup_steps=self.cfg.num_train_iter * 0)
            ## set SGD and cosine lr
            model.set_optimizer(optimizer, scheduler)

            if self.cfg.gpu is not None:
                torch.cuda.set_device(self.cfg.gpu)
                model.model = model.model.cuda(self.cfg.gpu)
            else:
                model.model = torch.nn.DataParallel(model.model).cuda()

            model.ema_model = copy.deepcopy(model.model)

            loader_dict = {}
            dset_dict = {"train_lb": train_lb_dset, "val_lb": val_lb_dset}

            loader_dict["train_lb"] = get_data_loader(
                dset_dict["train_lb"],
                self.cfg.batch_size,
                data_sampler=self.cfg.train_sampler,
                num_iters=self.cfg.num_train_iter,
                num_workers=self.cfg.num_workers,
            )
            loader_dict["val_lb"] = get_data_loader(
                dset_dict["val_lb"], self.cfg.eval_batch_size, num_workers=self.cfg.num_workers, drop_last=False
            )

            ## set DataLoader
            model.set_dataset(dset_dict)
            model.set_data_loader(loader_dict)

            save_model_path = Path(self.cfg.save_path) / "models" / f"model_{idx}.pt"
            save_val_logits_path = Path(self.cfg.save_path) / "models" / f"val_logits_{idx}.pt"
            
            if save_model_path.exists() and save_val_logits_path.exists():
                print(f"found existing model at {save_model_path}")
                print(f"found existing validation logits at {save_val_logits_path}, skip training.\n")
            else:
                # START TRAINING
                trainer = model.train
                trainer(self.cfg)
                if self.cfg.save_model:
                    model.save_model(save_model_path)
                
                raw_val_outputs, val_targets = model.get_logits(loader_dict["val_lb"], args=self.cfg)
                logits = {"val_logits": raw_val_outputs, "val_targets": val_targets}
                if self.cfg.save_model:
                    torch.save(logits, save_val_logits_path)

        random.setstate(random_state)
        np.random.set_state(np_random_state)
        torch.random.set_rng_state(torch_random_state)
        logging.warning("Training is FINISHED")

    def pre_check(self):
        if not torch.cuda.is_available():
            raise Exception("ONLY GPU TRAINING IS SUPPORTED")

        if self.cfg.gpu is not None:
            warnings.warn("You have chosen a specific GPU. This will completely disable data parallelism.")

    def estimate(self, ulb_data, ulb_targets=None, num_classes=None, test_transform=None, save_name=None, verbose=False):
        test_transform = test_transform if test_transform is not None else transforms.ToTensor()

        if num_classes is not None:
            self.cfg.num_classes = num_classes
        assert ulb_targets is None or self.cfg.num_classes > 0, "please provide the number of classes"
        
        # Construct Dataset & DataLoader
        ulb_dset = BasicDataset(ulb_data, ulb_targets, self.cfg.num_classes, test_transform, is_ulb=True, onehot=False)

        logits = self.get_logits(ulb_dset)
        estimations = self.apply_lse(logits)

        ulb_dist = None
        if ulb_targets is not None:
            ulb_dist = labels_to_dist(ulb_targets, self.cfg.num_classes)
            
        with np.printoptions(precision=3, suppress=True, formatter={"float": "{: 0.3f}".format}):
            if verbose:
                print(f"\nTarget distribution: {ulb_dist}")
                print("\nEstimations:")
            
            mse = None
            for name, est_target_dist in estimations.items():
                if ulb_dist is not None:
                    mse = mean_squared_error(ulb_dist, est_target_dist)
                    estimations[name] = {"estimation": est_target_dist.tolist(), "mse": mse}
                    if verbose: print(f"{name}: {est_target_dist}, MSE: {mse:.5f}")
                else:
                    estimations[name] = {"estimation": est_target_dist.tolist()}
                    if verbose: print(f"{name}: {est_target_dist}")
        
        if save_name is not None:
            save_est_path = Path(self.cfg.save_path) / f"{save_name}.json"
            with open(save_est_path, "w") as f:
                json.dump(estimations, f, indent=4)
            if verbose: logging.warning(f"Estimation Saved Successfully: {save_est_path}")

        return estimations
    
    def get_logits(self, ulb_dset):
        _net_builder = net_builder(
            self.cfg.net,
            self.cfg.net_from_name,
            {
                "first_stride": 2 if "stl" in self.cfg.dataset else 1,
                "depth": self.cfg.depth,
                "widen_factor": self.cfg.widen_factor,
                "leaky_slope": self.cfg.leaky_slope,
                "bn_momentum": self.cfg.bn_momentum,
                "dropRate": self.cfg.dropout,
                "use_embed": False,
            },
            verbose=False
        )

        logits = {"val_logits": [], "val_targets": [], "ulb_logits": []}
        ulb_loader = get_data_loader(ulb_dset, self.cfg.eval_batch_size, num_workers=self.cfg.num_workers, drop_last=False)
        model = ModelTrainer(_net_builder, self.cfg.num_classes, num_eval_iter=self.cfg.num_eval_iter, ema_m=self.cfg.ema_m)
        
        if self.cfg.gpu is not None:
            torch.cuda.set_device(self.cfg.gpu)
            model.model = model.model.cuda(self.cfg.gpu)
        else:
            model.model = torch.nn.DataParallel(model.model).cuda()

        model.ema_model = copy.deepcopy(model.model)
        
        for idx in range(self.cfg.num_ensemble):
            val_logits_path = Path(self.cfg.save_path) / "models" / f"val_logits_{idx}.pt"
            if val_logits_path.exists():
                data = torch.load(val_logits_path)
                logits["val_logits"].append(data["val_logits"])
                logits["val_targets"].append(data["val_targets"])
            else:
                raise RuntimeError(f"Can't find validation logits at {val_logits_path}, please run train_models() first.")
            
            model_path = Path(self.cfg.save_path) / "models" / f"model_{idx}.pt"
            if model_path.exists():
                model.load_model(model_path)
            else:
                raise RuntimeError(f"Can't find model at {model_path}, please run train_models() first.")
            
            raw_outputs, _ = model.get_logits(ulb_loader, args=self.cfg)
            logits["ulb_logits"].append(raw_outputs)
        
        return logits

    def apply_lse(self, logits):
        if self.cfg.lse_algs is None:
            raise RuntimeError("please provide a lse algorithm")
        
        assert len(logits["ulb_logits"]) == len(logits["val_logits"]) == len(logits["val_targets"]) >= self.cfg.num_ensemble
        ulb_logits = logits["ulb_logits"][: self.cfg.num_ensemble]
        val_logits = logits["val_logits"][: self.cfg.num_ensemble]
        val_targets = logits["val_targets"][: self.cfg.num_ensemble]
        
        estimations = {}
        names, estimators = get_lse_methods(self.cfg.lse_algs, self.cfg.calibrations, use_ensemble=True)
        for name, estimator in zip(names, estimators):
            estimator.fit(ulb_logits, val_logits, val_targets)
            estimations[name] = estimator.estim_target_dist

        return estimations



from dataclasses import dataclass, field


@dataclass
class Config:
    save_path: str = "./saved_models/temp"
    save_model: bool = True
    overwrite: bool = True
    num_train_iter: int = 16000
    num_eval_iter: int = 1600
    num_val_per_class: int = 5
    batch_size: int = 64
    eval_batch_size: int = 1024
    ema_m: float = 0.999
    use_mixup_drw: bool = True
    drw_warm: float = 0.75
    optim: str = "SGD"
    lr: float = 0.03
    momentum: float = 0.9
    weight_decay: float = 0.0005
    bn_momentum: float = 1.0 - 0.999
    amp: bool = False
    clip: float = 0
    net: str = "WideResNet"
    net_from_name: bool = False
    depth: int = 28
    widen_factor: int = 2
    leaky_slope: float = 0.1
    dropout: float = 0.0
    dataset: str = "custom"
    train_sampler: str = "RandomSampler"
    num_classes: int = 0
    num_workers: int = 1
    seed: int = 0
    gpu: int = 0
    lse_algs: list = field(default_factory=lambda: ["MLLS"])
    calibrations: list = field(default_factory=lambda: ["BCTS"])
    num_ensemble: int = 10


if __name__ == "__main__":
    config = Config(lse_algs=["BBSE", "RLLS", "MLLS"])
    print(config.num_ensemble)
    print(config.lse_algs)
    print(config)



import math
import yaml
import numpy as np

import torch
from torch.optim.lr_scheduler import LambdaLR


def softmax(logits):
    """
    Compute softmax values for each sets of scores in logits.

    Parameters:
        logits (numpy.ndarray): array containing m samples with n-dimensions (m,n)
    Returns:
        x_softmax (numpy.ndarray) softmaxed values for initial (m,n) array
    """
    e_logits = np.exp(logits - np.max(logits, axis=-1, keepdims=True))
    return e_logits / e_logits.sum(axis=-1, keepdims=True)


def inverse_softmax(softmax_preds):
    return np.log(softmax_preds) - np.mean(np.log(softmax_preds), axis=1)[:, None]


def to_one_hot(labels: np.array) -> np.array:
    if labels.ndim == 1:  # skip if labels are already one hot encoded
        n_labels = np.max(labels) + 1
        labels = np.eye(n_labels)[labels]
    return labels


def to_hard_labels(soft_labels, one_hot=True):
    hard_labels = np.argmax(soft_labels, axis=1)
    if one_hot:
        hard_labels = to_one_hot(hard_labels)
    return hard_labels


def get_confusion_matrix(labels, logits, normalize=False):
    assert logits.shape[0] == labels.shape[0]
    if labels.ndim != 1:  # check if labels are one hot encoded
        labels = np.argmax(labels, axis=1)

    num_class = np.unique(labels).shape[0]
    assert num_class == np.max(labels) + 1
    cf_mat = np.zeros((num_class, num_class))

    for logit, label in zip(logits, labels):
        cf_mat[:, label] += logit

    if normalize:
        # for c in range(num_class):
        #     cf_mat[:, c] /= cf_mat[:, c].sum()
        cf_mat /= cf_mat.sum()

    return cf_mat


def labels_to_dist(targets, num_classes, to_prob=True):
    ulb_dist = np.zeros(num_classes, dtype=int)
    for y in targets:
        ulb_dist[y] += 1
    if to_prob:
        ulb_dist = ulb_dist / ulb_dist.sum()
    return ulb_dist


def over_write_args_from_dict(args, dic):
    for k, v in dic.items():
        setattr(args, k, v)
    return args


def over_write_args_from_file(args, yml):
    if yml == "":
        return
    with open(yml, "r", encoding="utf-8") as f:
        dic = yaml.load(f.read(), Loader=yaml.Loader)
        for k in dic:
            setattr(args, k, dic[k])


def setattr_cls_from_kwargs(cls, kwargs, verbose=True):
    # if default values are in the cls,
    # overlap the value by kwargs
    for key in kwargs.keys():
        if hasattr(cls, key) and verbose:
            print(f"{key} in {cls} is overlapped by kwargs: {getattr(cls, key)} -> {kwargs[key]}")
        setattr(cls, key, kwargs[key])


def test_setattr_cls_from_kwargs():
    class _test_cls:
        def __init__(self):
            self.a = 1
            self.b = "hello"

    test_cls = _test_cls()
    config = {"a": 3, "b": "change_hello", "c": 5}
    setattr_cls_from_kwargs(test_cls, config)
    for key in config.keys():
        print(f"{key}:\t {getattr(test_cls, key)}")


def net_builder(net_name, from_name: bool, net_conf=None, is_remix=False, verbose=True):
    """
    return **class** of backbone network (not instance).
    Args
        net_name: 'WideResNet' or network names in torchvision.models
        from_name: If True, net_buidler takes models in torch.vision models. Then, net_conf is ignored.
        net_conf: When from_name is False, net_conf is the configuration of backbone network (now, only WRN is supported).
    """
    if from_name:
        import torchvision.models as models

        model_name_list = sorted(
            name for name in models.__dict__ if name.islower() and not name.startswith("__") and callable(models.__dict__[name])
        )

        if net_name not in model_name_list:
            assert Exception(
                f"[!] Networks' Name is wrong, check net config, \
                               expected: {model_name_list}  \
                               received: {net_name}"
            )
        else:
            return models.__dict__[net_name]

    else:
        if net_name == "WideResNet":
            import labelshift.model.nets.wrn as net

            builder = getattr(net, "build_WideResNet")()
        elif net_name == "ResNet50":
            import labelshift.model.nets.resnet50 as net

            builder = getattr(net, "build_ResNet50")(is_remix)
        else:
            assert Exception("Not Implemented Error")

        if net_name != "ResNet50":
            setattr_cls_from_kwargs(builder, net_conf, verbose=verbose)
        
        return builder.build


def test_net_builder(net_name, from_name, net_conf=None):
    builder = net_builder(net_name, from_name, net_conf)
    print(f"net_name: {net_name}, from_name: {from_name}, net_conf: {net_conf}")
    print(builder)


def get_optimizer(net, optim_name="SGD", lr=0.1, momentum=0.9, weight_decay=0.0, nesterov=True, bn_wd_skip=True):
    """
    return optimizer (name) in torch.optim.
    If bn_wd_skip, the optimizer does not apply
    weight decay regularization on parameters in batch normalization.
    """

    decay = []
    no_decay = []
    for name, param in net.named_parameters():
        if ("bn" in name or "bias" in name) and bn_wd_skip:
            no_decay.append(param)
        else:
            decay.append(param)

    per_param_args = [{"params": decay}, {"params": no_decay, "weight_decay": 0.0}]

    if optim_name == "SGD":
        optimizer = torch.optim.SGD(per_param_args, lr=lr, momentum=momentum, weight_decay=weight_decay, nesterov=nesterov)
    elif optim_name == "AdamW":
        optimizer = torch.optim.AdamW(per_param_args, lr=lr, weight_decay=weight_decay)
    elif optim_name == "Adam":
        optimizer = torch.optim.Adam(per_param_args, lr=lr)
    return optimizer


def get_cosine_schedule_with_warmup(optimizer, num_training_steps, num_cycles=7.0 / 16.0, num_warmup_steps=0, last_epoch=-1):
    """
    Get cosine scheduler (LambdaLR).
    if warmup is needed, set num_warmup_steps (int) > 0.
    """

    def _lr_lambda(current_step):
        """
        _lr_lambda returns a multiplicative factor given an interger parameter epochs.
        Decaying criteria: last_epoch
        """

        if current_step < num_warmup_steps:
            _lr = float(current_step) / float(max(1, num_warmup_steps))
        else:
            num_cos_steps = float(current_step - num_warmup_steps)
            num_cos_steps = num_cos_steps / float(max(1, num_training_steps - num_warmup_steps))
            _lr = max(0.0, math.cos(math.pi * num_cycles * num_cos_steps))
        return _lr

    return LambdaLR(optimizer, _lr_lambda, last_epoch)



import numpy as np

from labelshift.utils import to_one_hot, softmax, to_hard_labels

from .base import BaseEstimator


class MLLS(BaseEstimator):
    def __init__(self, hard_label=False, calibrator=None, tolerance=1e-6, max_iterations=100, verbose=False):
        self.verbose = verbose
        self.tolerance = tolerance
        self.hard_label = hard_label
        self.calibrator = calibrator
        self.max_iterations = max_iterations
        self.estim_source_dist = None  # estimate source class distribution
        self.estim_target_dist = None  # estimate target class distribution
        self.imp_weights = None  # importance weights

    def __str__(self):
        return "Maximum Likelihood Label Shift (MLLS)"

    def __copy__(self):
        calirator_copy = self.calibrator.__copy__() if self.calibrator is not None else None
        mlls_copy = MLLS(self.hard_label, calirator_copy, self.tolerance, self.max_iterations, self.verbose)
        mlls_copy.estim_source_dist = self.estim_source_dist
        mlls_copy.estim_target_dist = self.estim_target_dist
        mlls_copy.imp_weights = self.imp_weights
        return mlls_copy

    # TODO: add torch.tensor type support for target_logits, src_logits, src_labels
    def fit(self, target_logits, src_logits, src_labels=None):
        if src_labels is not None:
            src_labels = to_one_hot(src_labels)

        if self.calibrator is not None:
            assert src_labels is not None, "source_labels can't be None"
            self.calibrator.fit(src_logits, src_labels)
            src_logits = self.calibrator.calibrate(src_logits)
            target_logits = self.calibrator.calibrate(target_logits)
        else:
            src_logits = softmax(src_logits)
            target_logits = softmax(target_logits)

        if self.hard_label:
            target_logits = to_hard_labels(target_logits, one_hot=True)
            src_logits = to_hard_labels(src_logits, one_hot=True)

        est_src_dist = np.mean(src_labels, axis=0)
        est_src_dist = est_src_dist / np.sum(est_src_dist)  # normalize

        est_target_dist = est_src_dist
        for _ in range(self.max_iterations):
            est_target_posterior_probs = target_logits * (est_target_dist / est_src_dist)[None, :]
            est_target_posterior_probs = est_target_posterior_probs / np.sum(est_target_posterior_probs, axis=-1)[:, None]
            next_est_target_dist = np.mean(est_target_posterior_probs, axis=0)
            if np.all(np.abs(next_est_target_dist - est_target_dist) <= self.tolerance):
                break
            est_target_dist = next_est_target_dist

        self.estim_source_dist = est_src_dist
        self.estim_target_dist = next_est_target_dist
        self.imp_weights = self.estim_target_dist / self.estim_source_dist

        if self.verbose:
            with np.printoptions(precision=3, suppress=True, formatter={"float": "{: 0.3f}".format}):
                print(f"Est Source Class Distribution: {self.estim_source_dist}")
                print(f"Est Target Class Distribution: {self.estim_target_dist}")
                print(f"Est Importance Weights       : {self.imp_weights}")

        return self



import numpy as np

import torchvision
from torchvision import transforms

from .dataset import BasicDataset, ResampleDataset, ResampleDataset
from .data_utils import gen_imb_list, sample_data


mean, std = {}, {}
mean["cifar10"] = [x / 255 for x in [125.3, 123.0, 113.9]]
mean["cifar100"] = [x / 255 for x in [129.3, 124.1, 112.4]]

std["cifar10"] = [x / 255 for x in [63.0, 62.1, 66.7]]
std["cifar100"] = [x / 255 for x in [68.2, 65.4, 70.4]]


def get_transform(mean, std, crop_size, train=True):
    if train:
        return transforms.Compose(
            [
                transforms.RandomHorizontalFlip(),
                transforms.RandomCrop(crop_size, padding=4, padding_mode="reflect"),
                transforms.ToTensor(),
                transforms.Normalize(mean, std),
            ]
        )
    else:
        return transforms.Compose([transforms.ToTensor(), transforms.Normalize(mean, std)])


class Imbalanced_Dataset:
    """
    Imbalanced_Dataset class gets dataset from torchvision.datasets,
    separates labeled and unlabeled data,
    and return BasicDataset: torch.utils.data.Dataset (see datasets.dataset.py)
    """

    def __init__(self, name="cifar10", num_classes=10, data_dir="./data"):
        """
        Args
            name: name of dataset in torchvision.datasets (cifar10, cifar100, svhn, stl10)
            train: True means the dataset is training dataset (default=True)
            num_classes: number of label classes
            data_dir: path of directory, where data is downloaed or stored.
        """
        self.name = name
        self.train = True
        self.num_classes = num_classes
        self.data_dir = data_dir
        self.crop_size = 32
        self.train_transform = get_transform(mean[name], std[name], self.crop_size, True)
        self.test_transform = get_transform(mean[name], std[name], self.crop_size, False)

    def get_data(self):
        """
        get_data returns data (images) and targets (labels)
        shape of data: B, H, W, C
        shape of labels: B,
        """
        dset = getattr(torchvision.datasets, self.name.upper())
        if "CIFAR" in self.name.upper():
            dset = dset(self.data_dir, train=self.train, download=True)
            data, targets = dset.data, dset.targets
            return data, targets

    def get_lb_ulb_dset(
        self,
        max_labeled_per_class,
        max_unlabeled_per_class,
        lb_imb_ratio,
        ulb_imb_ratio,
        imb_type,
        onehot=False,
        seed=0,
    ):
        """
        get_lb_ulb_dset split training samples into labeled and unlabeled samples.
        The labeled and unlabeled data might be imbalanced over classes.

        Args:
            num_labels: number of labeled data.
            lb_img_ratio: imbalance ratio of labeled data.
            ulb_imb_ratio: imbalance ratio of unlabeled data.
            imb_type: type of imbalance data.
            onehot: If True, the target is converted into onehot vector.
            seed: Get deterministic results of labeled and unlabeld data.

        Returns:
            ResampleDataset (for labeled data), BasicDataset (for unlabeld data)
        """
        data, targets = self.get_data()

        state = np.random.get_state()
        np.random.seed(seed)

        data, target = np.array(data), np.array(targets)
        labeled_class_num_list = gen_imb_list(max_labeled_per_class, lb_imb_ratio, imb_type, self.num_classes)
        lb_data, lb_targets, lb_idx = sample_data(data, target, labeled_class_num_list, replace=False)

        rst_idx = np.array(sorted(list(set(range(len(data))) - set(lb_idx))))  # unlabeled_data index of data

        unlabeled_class_num_list = gen_imb_list(max_unlabeled_per_class, ulb_imb_ratio, imb_type, self.num_classes)
        ulb_data, ulb_targets, ulb_idx = sample_data(data[rst_idx], target[rst_idx], unlabeled_class_num_list, replace=False)
        ulb_idx = rst_idx[ulb_idx]  # correct the ulb_idx

        all_idx = np.concatenate([lb_idx, ulb_idx])
        assert np.unique(all_idx).shape == all_idx.shape  # check no duplicate value

        print(f"#labeled  : {labeled_class_num_list.sum()}, {labeled_class_num_list}")
        print(f"#unlabeled: {unlabeled_class_num_list.sum()}, {unlabeled_class_num_list}")

        np.random.set_state(state)

        lb_dset = ResampleDataset(lb_data, lb_targets, self.num_classes, self.train_transform, self.test_transform, onehot=onehot)
        ulb_dset = BasicDataset(ulb_data, ulb_targets, self.num_classes, self.test_transform, is_ulb=True, onehot=onehot)

        return lb_dset, ulb_dset



import numpy as np
from numpy.linalg import inv

from labelshift.utils import to_one_hot, softmax, to_hard_labels, get_confusion_matrix

from .base import BaseEstimator


class BBSE(BaseEstimator):
    def __init__(self, hard_label=False, calibrator=None, verbose=False):
        self.hard_label = hard_label
        self.calibrator = calibrator
        self.verbose = verbose
        self.estim_source_dist = None  # estimate source class distribution
        self.estim_target_dist = None  # estimate target class distribution
        self.imp_weights = None  # importance weights

    def __str__(self):
        return "Black Box Shift Estimation (BBSE)"

    def __copy__(self):
        calirator_copy = self.calibrator.__copy__() if self.calibrator is not None else None
        bbse_copy = BBSE(self.hard_label, calirator_copy, self.verbose)
        bbse_copy.estim_source_dist = self.estim_source_dist
        bbse_copy.estim_target_dist = self.estim_target_dist
        bbse_copy.imp_weights = self.imp_weights
        return bbse_copy

    def fit(self, target_logits, src_logits, src_labels):
        src_labels = to_one_hot(src_labels)

        if self.calibrator is not None:
            assert src_labels is not None, "source_labels can't be None"
            self.calibrator.fit(src_logits, src_labels)
            src_logits = self.calibrator.calibrate(src_logits)
            target_logits = self.calibrator.calibrate(target_logits)
        else:
            src_logits = softmax(src_logits)
            target_logits = softmax(target_logits)

        if self.hard_label:
            target_logits = to_hard_labels(target_logits, one_hot=True)
            src_logits = to_hard_labels(src_logits, one_hot=True)

        self.estim_source_dist = np.mean(src_labels, axis=0)
        self.estim_source_dist /= np.sum(self.estim_source_dist)  # normalize

        mean_target_logits = np.mean(target_logits, axis=0)
        confusion_matrix = get_confusion_matrix(src_labels, src_logits, normalize=True)

        self.imp_weights = np.matmul(inv(confusion_matrix), mean_target_logits)
        self.imp_weights = self.imp_weights * (self.imp_weights >= 0)  # mask out negative weights

        self.estim_target_dist = self.imp_weights * self.estim_source_dist
        self.estim_target_dist = self.estim_target_dist / np.sum(self.estim_target_dist)
        self.imp_weights = self.estim_target_dist / self.estim_source_dist

        if self.verbose:
            with np.printoptions(precision=3, suppress=True, formatter={"float": "{: 0.3f}".format}):
                print(f"Est Source Class Distribution: {self.estim_source_dist}")
                print(f"Est Target Class Distribution: {self.estim_target_dist}")
                print(f"Est Importance Weights       : {self.imp_weights}")

        return self



import numpy as np
from PIL import Image
from numpy.random import choice

from torch.utils.data import Dataset

from .data_utils import get_onehot


class BasicDataset(Dataset):
    """
    BasicDataset returns a pair of image and labels (targets).
    If targets are not given, BasicDataset returns None as the label.
    """

    def __init__(self, data, targets=None, num_classes=None, transform=None, is_ulb=False, onehot=False, *args, **kwargs):
        """
        Args
            data: x_data
            targets: y_data (if not exist, None)
            num_classes: number of label classes
            transform: basic transformation of data
            onehot: If True, label is converted into onehot vector.
        """
        super(BasicDataset, self).__init__()
        self.data = data
        self.targets = targets
        self.num_classes = num_classes
        self.transform = transform
        self.is_ulb = is_ulb
        self.onehot = onehot

    def __getitem__(self, idx):
        # set idx-th target
        if self.targets is None:
            target = None
        else:
            target_ = self.targets[idx]
            target = target_ if not self.onehot else get_onehot(self.num_classes, target_)

        img = self.data[idx]

        if self.transform is not None:
            if isinstance(img, np.ndarray):
                img = Image.fromarray(img)
            img = self.transform(img)

        if self.is_ulb:
            return img, -1
            # return img, target
        else:
            return img, target

    def __len__(self):
        return len(self.data)


class ResampleDataset(BasicDataset):
    """
    Random resampling subsets of a dataset.
    """

    def __init__(self, data, targets=None, num_classes=None, train_transform=None, test_transform=None, onehot=False):
        """
        Args
            data: x_data
            targets: y_data (if not exist, None)
            num_classes: number of label classes
            train_transform: transformation of training data
            test_transform: transformation of validation data
            onehot: If True, label is converted into onehot vector.
        """
        super().__init__(data, targets, num_classes, None, onehot)
        self.train_transform = train_transform
        self.test_transform = test_transform

    def resample(self, num_val_per_class, num_splits, replace=False, seed=0):
        state = np.random.get_state()
        np.random.seed(seed)

        n_idxes = []
        for _ in range(num_splits):
            val_idxes, train_idxes = [], []
            for c in range(self.num_classes):
                idxes = np.where(self.targets == c)[0]
                _val_idxes = choice(idxes, num_val_per_class, replace=replace)
                val_idxes.append(_val_idxes)
                _train_idxes = list(set(idxes) - set(_val_idxes))
                train_idxes.append(_train_idxes)
            val_idxes = np.hstack(val_idxes)
            train_idxes = np.hstack(train_idxes)
            assert not np.isin(val_idxes, train_idxes).any()  # check no overlap
            n_idxes.append((train_idxes, val_idxes))

        np.random.set_state(state)

        # split to training and validation dataset
        for idx, (train_idxes, val_idxes) in enumerate(n_idxes):
            train_lb_dset = Subset(self, train_idxes, self.train_transform)
            val_lb_dset = Subset(self, val_idxes, self.test_transform)
            yield idx, train_lb_dset, val_lb_dset


class Subset(Dataset):
    """
    Subset of a dataset at specified indices.

    Arguments:
        dataset (Dataset): The whole Dataset
        indices (sequence): Indices in the whole set selected for subset
    """

    def __init__(self, dataset, indices, transform):
        self.dataset = dataset
        self.indices = indices
        self.transform = transform

    def __getitem__(self, idx):
        img, target = self.dataset[self.indices[idx]]
        if self.transform is not None:
            if isinstance(img, np.ndarray):
                img = Image.fromarray(img)
            img = self.transform(img)
        return img, target

    def __len__(self):
        return len(self.indices)



# copyright: https://github.com/pytorch/pytorch/blob/master/torch/utils/data/sampler.py
# modified to let num_samples can be assigned even when replacement=False

import torch
from torch.utils.data import Sampler
# from torch.utils.data import Sampler, Sized, Optional, Iterator
from typing import Iterator, Iterable, Optional, Sequence, List, TypeVar, Generic, Sized, Union

class RandomSampler(Sampler[int]):
    r"""Samples elements randomly. If without replacement, then sample from a shuffled dataset.
    If with replacement, then user can specify :attr:`num_samples` to draw.
    Args:
        data_source (Dataset): dataset to sample from
        replacement (bool): samples are drawn on-demand with replacement if ``True``, default=``False``
        num_samples (int): number of samples to draw, default=`len(dataset)`. This argument
            is supposed to be specified only when `replacement` is ``True``.
        generator (Generator): Generator used in sampling.
    """
    data_source: Sized
    replacement: bool

    def __init__(self, data_source: Sized, replacement: bool = False,
                 num_samples: Optional[int] = None, generator=None) -> None:
        self.data_source = data_source
        self.replacement = replacement
        self._num_samples = num_samples
        self.generator = generator

        if not isinstance(self.replacement, bool):
            raise TypeError("replacement should be a boolean value, but got "
                            "replacement={}".format(self.replacement))

        # if self._num_samples is not None and not replacement:
        #     raise ValueError("With replacement=False, num_samples should not be specified, "
        #                      "since a random permute will be performed.")

        if not isinstance(self.num_samples, int) or self.num_samples <= 0:
            raise ValueError("num_samples should be a positive integer "
                             "value, but got num_samples={}".format(self.num_samples))

    @property
    def num_samples(self) -> int:
        # dataset size might change at runtime
        if self._num_samples is None:
            return len(self.data_source)
        return self._num_samples

    def __iter__(self) -> Iterator[int]:
        n = len(self.data_source)
        if self.generator is None:
            seed = int(torch.empty((), dtype=torch.int64).random_().item())
            generator = torch.Generator()
            generator.manual_seed(seed)
        else:
            generator = self.generator

        if self.replacement:
            for _ in range(self.num_samples // 32):
                yield from torch.randint(high=n, size=(32,), dtype=torch.int64, generator=generator).tolist()
            yield from torch.randint(high=n, size=(self.num_samples % 32,), dtype=torch.int64, generator=generator).tolist()
        else:
            for _ in range(self.num_samples // n):
                yield from torch.randperm(n, generator=generator).tolist()
            yield from torch.randperm(n, generator=generator).tolist()[:self.num_samples % n]

    def __len__(self) -> int:
        return self.num_samples


import contextlib
import numpy as np
from tqdm import tqdm
from pathlib import Path
from copy import deepcopy
from sklearn.metrics import *

import torch
import torch.nn.functional as F
from torch.cuda.amp import autocast, GradScaler

from .trainer_utils import DRWLoss, mixup_one_target, to_one_hot, EMA, Bn_Controller, ce_loss


class ModelTrainer:
    def __init__(self, net_builder, num_classes, num_eval_iter=1000, ema_m=0.999):
        """
        class DistributionEstimate contains setter of data_loader, optimizer, and model update methods.
        Args:
            net_builder: backbone network class (see net_builder in utils.py)
            num_classes: # of label classes
            num_eval_iter: frequency of evaluation.
        """
        super(ModelTrainer, self).__init__()

        # momentum update param
        self.loader = {}
        self.num_classes = num_classes

        # create the encoders
        # network is builded only by num_classes,
        # other configs are covered in main.py

        self.model = net_builder(num_classes=num_classes)
        self.num_eval_iter = num_eval_iter

        self.optimizer = None
        self.scheduler = None

        self.it = 0

        self.ema_m = ema_m
        self.ema_model = deepcopy(self.model)

        self.bn_controller = Bn_Controller()

        self.checkpoint = None

    def set_data_loader(self, loader_dict):
        self.loader_dict = loader_dict
        print(f"[!] data loader keys: {self.loader_dict.keys()}")

    def set_dataset(self, dset_dict):
        self.dset_dict = dset_dict
        print(f"[!] datasets keys: {self.dset_dict.keys()}")

    def set_optimizer(self, optimizer, scheduler=None):
        self.optimizer = optimizer
        self.scheduler = scheduler

    def train(self, args):

        # lb: labeled, ulb: unlabeled
        self.model.train()
        self.ema = EMA(self.model, self.ema_m)
        self.ema.register()

        best_eval_acc, best_it = 0.0, 0

        scaler = GradScaler()
        amp_cm = autocast if args.amp else contextlib.nullcontext

        # deferred reweighting loss
        lb_dist_np = np.zeros(args.num_classes)
        for _, y in self.dset_dict["train_lb"]:
            lb_dist_np[y] += 1
        drw_loss = DRWLoss(args, lb_dist_np, warm_iter=int(args.drw_warm * args.num_train_iter))

        assert args.num_train_iter >= self.num_eval_iter, "num_eval_iter is smaller than num_train_iter"

        # start training
        pbar = tqdm(total=self.num_eval_iter, desc="training", leave=False)
        for x_lb, y_lb in self.loader_dict["train_lb"]:

            # prevent the training iterations exceed args.num_train_iter
            if self.it > args.num_train_iter:
                break
            torch.cuda.synchronize()

            x_lb = x_lb.cuda(args.gpu)
            y_lb = y_lb.cuda(args.gpu)

            # inference and calculate sup/unsup losses
            with amp_cm():
                if args.use_mixup_drw:
                    y_lb = to_one_hot(y_lb, self.num_classes, args.gpu)
                    mixed_x, mixed_y, _ = mixup_one_target(x_lb, y_lb, args.gpu)
                    logits_x = self.model(mixed_x)
                    total_loss = drw_loss(self.it, logits_x, mixed_y, use_hard_labels=False)
                    total_loss = total_loss.mean()
                else:
                    logits_x = self.model(x_lb)
                    total_loss = ce_loss(logits_x, y_lb, reduction="mean")

            # parameter updates
            if args.amp:
                scaler.scale(total_loss).backward()
                if args.clip > 0:
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), args.clip)
                scaler.step(self.optimizer)
                scaler.update()
            else:
                total_loss.backward()
                if args.clip > 0:
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), args.clip)
                self.optimizer.step()

            self.scheduler.step()
            self.ema.update()
            self.model.zero_grad()

            torch.cuda.synchronize()

            self.it += 1
            pbar.update(1)

            if self.it % self.num_eval_iter == 0:
                tb_dict = {
                    "train/loss": total_loss.detach().cpu().item(),
                    "lr": self.optimizer.param_groups[0]["lr"],
                }
                eval_dict = self.evaluate(self.loader_dict["val_lb"], args=args)
                tb_dict.update(eval_dict)

                if tb_dict["eval/top-1-acc"] > best_eval_acc:
                    best_eval_acc = tb_dict["eval/top-1-acc"]
                    best_it = self.it

                print(f"\n[{self.it}/{args.num_train_iter}] BEST_EVAL_ACC: {best_eval_acc}, at {best_it} iters, {tb_dict}")

                del tb_dict
                pbar.reset()

        return eval_dict

    @torch.no_grad()
    def evaluate(self, eval_loader=None, args=None):
        self.model.eval()
        self.ema.apply_shadow()
        if eval_loader is None:
            eval_loader = self.loader_dict["val_lb"]
        total_loss = 0.0
        total_num = 0.0
        y_true = []
        y_pred = []
        y_logits = []
        for x, y in tqdm(eval_loader, desc="evaluating", leave=False):
            x, y = x.cuda(args.gpu), y.cuda(args.gpu)
            num_batch = x.shape[0]
            total_num += num_batch
            logits = self.model(x)
            loss = F.cross_entropy(logits, y, reduction="mean")
            y_true.extend(y.cpu().tolist())
            y_pred.extend(torch.max(logits, dim=-1)[1].cpu().tolist())
            y_logits.extend(torch.softmax(logits, dim=-1).cpu().tolist())
            total_loss += loss.detach() * num_batch
        top1 = accuracy_score(y_true, y_pred)
        top5 = top_k_accuracy_score(y_true, y_logits, k=5)
        self.ema.restore()
        self.model.train()
        return {"eval/loss": (total_loss / total_num).cpu().item(), "eval/top-1-acc": top1, "eval/top-5-acc": top5}

    @torch.no_grad()
    def get_logits(self, loader=None, args=None):
        self.model.eval()
        self.ema.apply_shadow()
        if loader is None:
            loader = self.loader_dict["val_lb"]

        y_true = []
        y_logits = []
        for x, y in tqdm(loader, desc="get logits", leave=False):
            x, y = x.cuda(args.gpu), y.cuda(args.gpu)
            logits = self.model(x)
            y_true.extend(y.cpu().tolist())
            y_logits.extend(logits.cpu().numpy())
        self.ema.restore()
        self.model.train()
        return np.array(y_logits), np.array(y_true)

    def save_model(self, save_path):
        save_path = Path(save_path)
        save_path.parent.mkdir(parents=True, exist_ok=True)

        # copy EMA parameters to ema_model for saving with model as temp
        self.model.eval()
        self.ema.apply_shadow()
        ema_model = deepcopy(self.model)
        self.ema.restore()
        self.model.train()

        torch.save(
            {
                "model": self.model.state_dict(),
                "optimizer": self.optimizer.state_dict(),
                "scheduler": self.scheduler.state_dict(),
                "it": self.it,
                "ema_model": ema_model.state_dict(),
            },
            save_path,
        )

        print(f"model saved: {save_path}")

    def load_model(self, load_path):
        checkpoint = torch.load(load_path)

        self.model.load_state_dict(checkpoint["model"])
        if self.optimizer is not None:
            self.optimizer.load_state_dict(checkpoint["optimizer"])
        if self.scheduler is not None:
            self.scheduler.load_state_dict(checkpoint["scheduler"])
        self.it = checkpoint["it"]
        self.ema_model.load_state_dict(checkpoint["ema_model"])

        self.ema = EMA(self.model, self.ema_m)
        self.ema.register()
        self.ema.load(self.ema_model)

        print(f"model loaded: {load_path}")


if __name__ == "__main__":
    pass



""" Setup
"""
from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="labelshift",
    version="0.1.12",
    license="MIT",
    description="A Pytorch-based toolbox for label shift estimation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pm25/Label-Shift-Estimation",
    author="Pin-Yen Huang",
    author_email="pyhuang97@gmail.com",
    # Note that this is a string of words separated by whitespace, not a list.
    keywords="pytorch label-shift-estimation",
    packages=find_packages(exclude=["data", "configs", "saved_models"]),
    include_package_data=True,
    install_requires=["torch >= 1.8", "torchvision", "transformers", "cvxpy"],
    python_requires=">=3.7",
)
